# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import io
import json
from datetime import datetime, timedelta

import pytest
from botocore.credentials import Credentials as StaticCredentials
from botocore.credentials import ReadOnlyCredentials, RefreshableCredentials
from botocore.session import Session
from dateutil.tz import tzutc

from awscli.customizations.configure.exportcreds import (
    BashEnvVarFormatter,
    BashNoExportEnvFormatter,
    ConfigureExportCredentialsCommand,
    CredentialProcessFormatter,
    Credentials,
    FishShellFormatter,
    PowershellFormatter,
    WindowsCmdFormatter,
    convert_botocore_credentials,
)
from awscli.customizations.exceptions import ConfigurationError
from awscli.testutils import mock, unittest


class JSONValue:
    def __init__(self, json_str):
        self.json_str = json_str
        self.parsed_json = json.loads(json_str)

    def __eq__(self, other):
        if isinstance(other, str):
            other_parsed = json.loads(other)
            return self.parsed_json == other_parsed
        return self.parsed_json == other


@pytest.mark.parametrize(
    'format_cls, expected',
    [
        (
            BashEnvVarFormatter,
            (
                (
                    'export AWS_ACCESS_KEY_ID=access_key\n'
                    'export AWS_SECRET_ACCESS_KEY=secret_key\n'
                ),
                (
                    'export AWS_ACCESS_KEY_ID=access_key\n'
                    'export AWS_SECRET_ACCESS_KEY=secret_key\n'
                    'export AWS_SESSION_TOKEN=token\n'
                    'export AWS_CREDENTIAL_EXPIRATION=2023-01-01T00:00:00Z\n'
                ),
            ),
        ),
        (
            BashNoExportEnvFormatter,
            (
                (
                    'AWS_ACCESS_KEY_ID=access_key\n'
                    'AWS_SECRET_ACCESS_KEY=secret_key\n'
                ),
                (
                    'AWS_ACCESS_KEY_ID=access_key\n'
                    'AWS_SECRET_ACCESS_KEY=secret_key\n'
                    'AWS_SESSION_TOKEN=token\n'
                    'AWS_CREDENTIAL_EXPIRATION=2023-01-01T00:00:00Z\n'
                ),
            ),
        ),
        (
            PowershellFormatter,
            (
                (
                    '$Env:AWS_ACCESS_KEY_ID="access_key"\n'
                    '$Env:AWS_SECRET_ACCESS_KEY="secret_key"\n'
                ),
                (
                    '$Env:AWS_ACCESS_KEY_ID="access_key"\n'
                    '$Env:AWS_SECRET_ACCESS_KEY="secret_key"\n'
                    '$Env:AWS_SESSION_TOKEN="token"\n'
                    '$Env:AWS_CREDENTIAL_EXPIRATION="2023-01-01T00:00:00Z"\n'
                ),
            ),
        ),
        (
            WindowsCmdFormatter,
            (
                (
                    'set AWS_ACCESS_KEY_ID=access_key\n'
                    'set AWS_SECRET_ACCESS_KEY=secret_key\n'
                ),
                (
                    'set AWS_ACCESS_KEY_ID=access_key\n'
                    'set AWS_SECRET_ACCESS_KEY=secret_key\n'
                    'set AWS_SESSION_TOKEN=token\n'
                    'set AWS_CREDENTIAL_EXPIRATION=2023-01-01T00:00:00Z\n'
                ),
            ),
        ),
        (
            FishShellFormatter,
            (
                (
                    'set -gx AWS_ACCESS_KEY_ID "access_key"\n'
                    'set -gx AWS_SECRET_ACCESS_KEY "secret_key"\n'
                ),
                (
                    'set -gx AWS_ACCESS_KEY_ID "access_key"\n'
                    'set -gx AWS_SECRET_ACCESS_KEY "secret_key"\n'
                    'set -gx AWS_SESSION_TOKEN "token"\n'
                    'set -gx AWS_CREDENTIAL_EXPIRATION "2023-01-01T00:00:00Z"\n'
                ),
            ),
        ),
        (
            CredentialProcessFormatter,
            (
                JSONValue(
                    '{"Version": 1, "AccessKeyId": "access_key", '
                    '"SecretAccessKey": "secret_key"}'
                ),
                JSONValue(
                    '{"Version": 1, "AccessKeyId": "access_key", '
                    '"SecretAccessKey": "secret_key", "SessionToken": '
                    '"token", "Expiration": "2023-01-01T00:00:00Z"}'
                ),
            ),
        ),
    ],
)
def test_cred_formatter(format_cls, expected):
    stream = io.StringIO()
    expected_static, expected_temporary = expected
    formatter = format_cls(stream)

    # Static case, only access_key/secret_key
    creds = Credentials('access_key', 'secret_key', None, None)
    formatter.display_credentials(creds)
    assert stream.getvalue() == expected_static

    # Refreshable case, access_key/secret_key/token/expiration.
    # Reset back to an empty stream.
    stream.truncate(0)
    stream.seek(0)
    expiry = '2023-01-01T00:00:00Z'
    temp_creds = Credentials('access_key', 'secret_key', 'token', expiry)
    formatter.display_credentials(temp_creds)
    assert stream.getvalue() == expected_temporary


class TestCanConvertBotocoreCredentials(unittest.TestCase):
    def test_can_convert_static_creds_with_no_expiry(self):
        self.assertEqual(
            convert_botocore_credentials(
                StaticCredentials('access_key', 'secret_key')
            ),
            Credentials(
                'access_key', 'secret_key', token=None, expiry_time=None
            ),
        )

    def test_can_convert_creds_with_token_and_no_expiry(self):
        self.assertEqual(
            convert_botocore_credentials(
                StaticCredentials('access_key', 'secret_key', 'token')
            ),
            Credentials('access_key', 'secret_key', 'token', expiry_time=None),
        )

    def test_can_convert_refreshable_with_expiry(self):
        expiry = datetime.now(tzutc()) + timedelta(hours=12)
        self.assertEqual(
            convert_botocore_credentials(
                RefreshableCredentials(
                    'access_key',
                    'secret_key',
                    'token',
                    expiry_time=expiry,
                    refresh_using=None,
                    method='explicit',
                )
            ),
            Credentials(
                'access_key',
                'secret_key',
                'token',
                expiry_time=expiry.isoformat(),
            ),
        )

    def test_no_expiry_time_if_non_datetime_value(self):
        bad_creds = mock.Mock(spec=StaticCredentials)
        bad_creds.get_frozen_credentials.return_value = ReadOnlyCredentials(
            'access_key', 'secret_key', 'token'
        )
        bad_creds._expiry_time = 'not a datetime'
        self.assertEqual(
            convert_botocore_credentials(bad_creds),
            Credentials('access_key', 'secret_key', 'token', expiry_time=None),
        )


class TestConfigureExportCredentialsCommand(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock(spec=Session)
        self.session.emit_first_non_none_response.return_value = None
        self.out_stream = io.StringIO()
        self.err_stream = io.StringIO()
        self.os_env = {}
        self.session.get_config_variable.return_value = 'default'
        self.export_creds_cmd = ConfigureExportCredentialsCommand(
            self.session, self.out_stream, self.err_stream, env=self.os_env
        )
        self.global_args = mock.Mock()
        self.expiry = '2023-01-01T00:00:00Z'
        self.creds = StaticCredentials('access_key', 'secret_key')

    def test_can_export_creds(self):
        self.session.get_credentials.return_value = self.creds
        rc = self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertEqual(rc, 0)
        self.assertEqual(
            self.out_stream.getvalue(),
            JSONValue(
                '{"Version": 1, "AccessKeyId": "access_key", '
                '"SecretAccessKey": "secret_key"}'
            ),
        )

    def test_can_export_creds_explicit_format(self):
        self.session.get_credentials.return_value = self.creds
        rc = self.export_creds_cmd(
            args=['--format', 'env'], parsed_globals=self.global_args
        )
        self.assertEqual(rc, 0)
        self.assertEqual(
            self.out_stream.getvalue(),
            'export AWS_ACCESS_KEY_ID=access_key\n'
            'export AWS_SECRET_ACCESS_KEY=secret_key\n',
        )

    def test_show_error_when_no_cred(self):
        self.session.get_credentials.return_value = None
        with pytest.raises(ConfigurationError) as excinfo:
            self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertIn('Unable to retrieve credentials', str(excinfo))

    def test_show_error_when_cred_resolution_errors(self):
        self.session.get_credentials.side_effect = Exception(
            "resolution failed"
        )
        with pytest.raises(ConfigurationError) as excinfo:
            self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertIn('resolution failed', str(excinfo))

    def test_can_detect_recursive_resolution(self):
        self.os_env['_AWS_CLI_PROFILE_CHAIN'] = 'default'
        with pytest.raises(ConfigurationError) as excinfo:
            self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertIn(
            'Credential process resolution detected an infinite loop',
            str(excinfo),
        )

    def test_nested_calls_not_recursive(self):
        self.session.get_credentials.return_value = self.creds
        # If we've seen the profiles foo and bar:
        self.os_env['_AWS_CLI_PROFILE_CHAIN'] = 'foo,bar'
        # And we're now on baz, then we shouldn't get an error
        # because there's no cycle.
        self.session.get_config_variable.return_value = 'baz'
        rc = self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertEqual(rc, 0)
        self.assertEqual(
            self.out_stream.getvalue(),
            JSONValue(
                '{"Version": 1, "AccessKeyId": "access_key", '
                '"SecretAccessKey": "secret_key"}'
            ),
        )

    def test_nested_calls_with_cycle(self):
        self.session.get_credentials.return_value = self.creds
        self.os_env['_AWS_CLI_PROFILE_CHAIN'] = 'foo,bar,baz'
        self.session.get_config_variable.return_value = 'bar'
        with pytest.raises(ConfigurationError) as excinfo:
            self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertIn(
            'Credential process resolution detected an infinite loop',
            str(excinfo),
        )

    def test_handles_comma_char_in_profile_name_no_cycle(self):
        self.session.get_credentials.return_value = self.creds
        self.os_env['_AWS_CLI_PROFILE_CHAIN'] = 'foo,bar,baz'
        # Shouldn't be a cycle, the comma is part of the profile name.
        self.session.get_config_variable.return_value = 'bar,baz'
        rc = self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertEqual(rc, 0)
        self.assertEqual(
            self.out_stream.getvalue(),
            JSONValue(
                '{"Version": 1, "AccessKeyId": "access_key", '
                '"SecretAccessKey": "secret_key"}'
            ),
        )

    def test_detects_comma_char_with_cycle(self):
        self.session.get_credentials.return_value = self.creds
        # To test the round tripping of CSV writing/reading, we'll invoke
        # the command twice to ensure however it records the profiles
        # that it detects a cycle with names that need to be escaped/quoted.
        self.os_env['_AWS_CLI_PROFILE_CHAIN'] = 'foo,bar'
        self.session.get_config_variable.return_value = 'bar,baz'
        # First time it succeeds.
        rc = self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertEqual(rc, 0)
        # Second time, it detects the cycle.
        second_invoke = ConfigureExportCredentialsCommand(
            self.session, self.out_stream, self.err_stream, env=self.os_env
        )
        with pytest.raises(ConfigurationError) as excinfo:
            second_invoke(args=[], parsed_globals=self.global_args)
        self.assertIn(
            'Credential process resolution detected an infinite loop',
            str(excinfo),
        )

    def test_max_recursion_limit(self):
        self.session.get_credentials.return_value = self.creds
        self.os_env['_AWS_CLI_PROFILE_CHAIN'] = ','.join(
            ['a', 'b', 'c', 'd', 'e', 'f', 'g']
        )
        with pytest.raises(ConfigurationError) as excinfo:
            self.export_creds_cmd(args=[], parsed_globals=self.global_args)
        self.assertIn(
            'Maximum recursive credential process resolution reached',
            str(excinfo),
        )



# Copyright 2022 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os

import botocore.session
import pytest


@pytest.fixture
def loader():
    return botocore.session.Session().get_component('data_loader')


def _available_services():
    return botocore.session.Session().get_available_services()


@pytest.mark.parametrize("service", _available_services())
def test_models_contain_only_known_file_types(service, loader):
    known_types = {
        "completions-1",
        "endpoint-rule-set-1",
        "paginators-1",
        "service-2",
        "waiters-2",
    }
    api_version = loader.determine_latest_version(service, "service-2")
    service_dir = os.path.join(loader.BUILTIN_DATA_PATH, service, api_version)
    for model_file in os.listdir(service_dir):
        assert (
            model_file.split(".")[0] in known_types
        ), f'Found unexpected model file: \'{model_file}\' in \'{service_dir}\''


@pytest.mark.parametrize("service", _available_services())
def test_models_contains_single_api_version_per_service(service, loader):
    assert len(loader.list_api_versions(service, 'service-2')) == 1


def test_models_contain_no_blocklisted_services():
    blocklisted_services = ['transcribe-streaming', 'sms-voice']
    services = _available_services()
    for service in blocklisted_services:
        assert service not in services


@pytest.mark.parametrize(
    ("service", "type_name"), [("autoscaling", "waiters-2")]
)
def test_service_contains_no_unexpected_file_type(service, type_name, loader):
    assert service not in loader.list_available_services(type_name)


def test_dbsnapshotcompleted_waiter_in_rds(loader):
    data = loader.load_service_model('rds', 'waiters-2')
    assert data['waiters'].get('DBSnapshotCompleted') is not None



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
import os
import signal
import sys

from awscli import (
    _DEFAULT_BASE_REMOTE_URL,
)
from awscli.argparser import HELP_BLURB, ArgParseException
from awscli.compat import StringIO
from awscli.help import (
    ExecutableNotFoundError,
    HelpCommand,
    PosixHelpRenderer,
    PosixPagingHelpRenderer,
    ProviderHelpCommand,
    TopicHelpCommand,
    TopicListerCommand,
    WindowsHelpRenderer,
    WindowsPagingHelpRenderer,
)
from awscli.testutils import FileCreator, mock, skip_if_windows, unittest


class HelpSpyMixin:
    def __init__(self):
        self.exists_on_path = {}
        self.popen_calls = []
        self.mock_popen = mock.Mock()

    def _exists_on_path(self, name):
        return self.exists_on_path.get(name)

    def _popen(self, *args, **kwargs):
        self.popen_calls.append((args, kwargs))
        return self.mock_popen


class FakePosixHelpRenderer(HelpSpyMixin, PosixHelpRenderer):
    def __init__(self, output_stream=sys.stdout):
        HelpSpyMixin.__init__(self)
        PosixHelpRenderer.__init__(self, output_stream)


class FakeWindowsHelpRenderer(HelpSpyMixin, WindowsHelpRenderer):
    def __init__(self, output_stream=sys.stdout):
        HelpSpyMixin.__init__(self)
        WindowsHelpRenderer.__init__(self, output_stream)


class TestHelpPager(unittest.TestCase):
    def setUp(self):
        self.environ = {}
        self.environ_patch = mock.patch('os.environ', self.environ)
        self.environ_patch.start()
        self.renderer = PosixHelpRenderer()

    def tearDown(self):
        self.environ_patch.stop()

    def test_no_env_vars(self):
        self.assertEqual(
            self.renderer.get_pager_cmdline(), self.renderer.PAGER.split()
        )

    def test_manpager(self):
        pager_cmd = 'foobar'
        os.environ['MANPAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(), pager_cmd.split())

    def test_pager(self):
        pager_cmd = 'fiebaz'
        os.environ['PAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(), pager_cmd.split())

    def test_both(self):
        os.environ['MANPAGER'] = 'foobar'
        os.environ['PAGER'] = 'fiebaz'
        self.assertEqual(self.renderer.get_pager_cmdline(), 'foobar'.split())

    def test_manpager_with_args(self):
        pager_cmd = 'less -X'
        os.environ['MANPAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(), pager_cmd.split())

    def test_pager_with_args(self):
        pager_cmd = 'less -X --clearscreen'
        os.environ['PAGER'] = pager_cmd
        self.assertEqual(self.renderer.get_pager_cmdline(), pager_cmd.split())

    @skip_if_windows('Requires posix system.')
    def test_no_groff_or_mandoc_exists(self):
        renderer = FakePosixHelpRenderer()
        renderer.exists_on_path['groff'] = False
        renderer.exists_on_path['mandoc'] = False
        expected_error = 'Could not find executable named "groff or mandoc"'
        with self.assertRaisesRegex(ExecutableNotFoundError, expected_error):
            renderer.render('foo')

    @skip_if_windows('Requires POSIX system.')
    def test_renderer_falls_back_to_mandoc(self):
        stdout = StringIO()
        renderer = FakePosixHelpRenderer(output_stream=stdout)

        renderer.exists_on_path['groff'] = False
        renderer.exists_on_path['mandoc'] = True
        renderer.mock_popen.communicate.return_value = (b'foo', '')
        renderer.render('foo')
        self.assertEqual(stdout.getvalue(), 'foo\n')

    @skip_if_windows('Requires POSIX system.')
    def test_no_pager_exists(self):
        fake_pager = 'foobar'
        os.environ['MANPAGER'] = fake_pager
        stdout = StringIO()
        renderer = FakePosixHelpRenderer(output_stream=stdout)
        renderer.exists_on_path[fake_pager] = False

        renderer.exists_on_path['groff'] = True
        renderer.mock_popen.communicate.return_value = (b'foo', '')
        renderer.render('foo')
        self.assertEqual(stdout.getvalue(), 'foo\n')

    def test_shlex_split_for_pager_var(self):
        pager_cmd = '/bin/sh -c "col -bx | vim -c \'set ft=man\' -"'
        os.environ['PAGER'] = pager_cmd
        self.assertEqual(
            self.renderer.get_pager_cmdline(),
            ['/bin/sh', '-c', "col -bx | vim -c 'set ft=man' -"],
        )

    def test_can_render_contents(self):
        renderer = FakePosixHelpRenderer()
        renderer.exists_on_path['groff'] = True
        renderer.exists_on_path['less'] = True
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('foo')
        self.assertEqual(renderer.popen_calls[-1][0], (['less', '-R'],))

    def test_can_page_output_on_windows(self):
        renderer = FakeWindowsHelpRenderer()
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('foo')
        self.assertEqual(renderer.popen_calls[-1][0], (['more'],))

    @skip_if_windows("Ctrl-C not valid on windows.")
    def test_can_handle_ctrl_c(self):
        class CtrlCRenderer(FakePosixHelpRenderer):
            def _popen(self, *args, **kwargs):
                if self._is_pager_call(args):
                    os.kill(os.getpid(), signal.SIGINT)
                return self.mock_popen

            def _is_pager_call(self, args):
                return 'less' in args[0]

        renderer = CtrlCRenderer()
        renderer.mock_popen.communicate.return_value = ('send to pager', '')
        renderer.exists_on_path['groff'] = True
        renderer.exists_on_path['less'] = True
        renderer.render('foo')
        last_call = renderer.mock_popen.communicate.call_args_list[-1]
        self.assertEqual(last_call, mock.call(input='send to pager'))

    def test_no_warnings_on_invalid_rst(self):
        renderer = FakePosixHelpRenderer()
        renderer.exists_on_path['groff'] = True
        renderer.exists_on_path['less'] = True
        renderer.mock_popen.communicate.return_value = ('rendered', '')
        renderer.render('*****foo')
        self.assertFalse(
            'WARNING' in renderer.mock_popen.communicate.call_args_list[0]
        )


class TestHelpCommandBase(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.file_creator = FileCreator()

    def tearDown(self):
        self.file_creator.remove_all()


class TestHelpCommand(TestHelpCommandBase):
    """Test some of the deeper functionality of the HelpCommand

    We do this by subclassing from HelpCommand and ensure it is behaving
    as expected.
    """

    def setUp(self):
        super().setUp()
        self.doc_handler_mock = mock.Mock()
        self.subcommand_mock = mock.Mock()
        self.renderer = mock.Mock()

        class SampleHelpCommand(HelpCommand):
            EventHandlerClass = self.doc_handler_mock

            @property
            def subcommand_table(sample_help_cmd_self):
                return {'mycommand': self.subcommand_mock}

        self.cmd = SampleHelpCommand(self.session, None, None, None)
        self.cmd.renderer = self.renderer

    def test_subcommand_call(self):
        self.cmd(['mycommand'], None)
        self.subcommand_mock.assert_called_with([], None)
        self.assertFalse(self.doc_handler_mock.called)

    def test_regular_call(self):
        self.cmd([], None)
        self.assertFalse(self.subcommand_mock.called)
        self.doc_handler_mock.assert_called_with(self.cmd)
        self.assertTrue(self.renderer.render.called)

    def test_invalid_subcommand(self):
        with self.assertRaises(ArgParseException) as e:
            self.cmd(['no-exist-command'], None)
        # We should see the pointer to "aws help" in the error message.
        self.assertIn(HELP_BLURB, str(e.exception))


class TestProviderHelpCommand(TestHelpCommandBase):
    def setUp(self):
        super().setUp()
        self.session.provider = None
        self.command_table = {}
        self.arg_table = {}
        self.description = None
        self.synopsis = None
        self.usage = None

        # Create a temporary index file for ``aws help [command]`` to use.
        self.tags_dict = {'topic-name-1': {}, 'topic-name-2': {}}
        json_index = self.file_creator.create_file('index.json', '')
        with open(json_index, 'w') as f:
            json.dump(self.tags_dict, f, indent=4, sort_keys=True)
        self.json_patch = mock.patch(
            'awscli.topictags.TopicTagDB.index_file', json_index
        )
        self.json_patch.start()

        self.cmd = ProviderHelpCommand(
            self.session,
            self.command_table,
            self.arg_table,
            self.description,
            self.synopsis,
            self.usage,
        )

    def tearDown(self):
        self.json_patch.stop()
        super().tearDown()

    def test_related_items(self):
        self.assertEqual(self.cmd.related_items, ['aws help topics'])

    def test_subcommand_table(self):
        subcommand_table = self.cmd.subcommand_table

        self.assertEqual(len(subcommand_table), 3)

        # Ensure there is a topics command
        self.assertIn('topics', subcommand_table)
        self.assertIsInstance(subcommand_table['topics'], TopicListerCommand)

        # Ensure the topics are there as well
        self.assertIn('topic-name-1', subcommand_table)
        self.assertIsInstance(
            subcommand_table['topic-name-1'], TopicHelpCommand
        )
        self.assertEqual(subcommand_table['topic-name-1'].name, 'topic-name-1')

        self.assertIn('topic-name-2', subcommand_table)
        self.assertIsInstance(
            subcommand_table['topic-name-2'], TopicHelpCommand
        )
        self.assertEqual(subcommand_table['topic-name-2'].name, 'topic-name-2')

    def test_url(self):
        self.assertIn(_DEFAULT_BASE_REMOTE_URL, self.cmd.url)
        self.assertIn("/index.html", self.cmd.url)


class TestTopicListerCommand(TestHelpCommandBase):
    def setUp(self):
        super().setUp()
        self.cmd = TopicListerCommand(self.session)

    def test_event_class(self):
        self.assertEqual(self.cmd.event_class, 'topics')

    def test_name(self):
        self.assertEqual(self.cmd.name, 'topics')


class TestTopicHelpCommand(TestHelpCommandBase):
    def setUp(self):
        super().setUp()
        self.name = 'topic-name-1'
        self.cmd = TopicHelpCommand(self.session, self.name)

    def test_event_class(self):
        self.assertEqual(self.cmd.event_class, 'topics.topic-name-1')

    def test_name(self):
        self.assertEqual(self.cmd.name, self.name)

    def test_url(self):
        self.assertIn(_DEFAULT_BASE_REMOTE_URL, self.cmd.url)
        self.assertIn("/topic/topic-name-1.html", self.cmd.url)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from unittest.mock import patch

from awscli.customizations.configure.sso_commands import ConfigureSSOCommand
from awscli.testutils import BaseAWSCommandParamsTest, mock
from tests.functional.sso import BaseSSOTest


class TestConfigureSSO(BaseAWSCommandParamsTest):
    prefix = 'configure sso'

    def test_sso_logout_do_not_exist(self):
        cmdline = self.prefix + ' logout'
        _, err, _ = self.run_cmd(cmdline, 252)
        self.assertIn('Unknown options', err)


class TestConfigureSSOCommand(BaseSSOTest):
    def setUp(self):
        super().setUp()

        self.registration_args_prompt = patch.object(
            ConfigureSSOCommand,
            '_prompt_for_sso_registration_args',
            return_value={
                'session_name': 'my-session',
                'sso_region': 'us-east-1',
                'start_url': 'https://identitycenter.amazonaws.com/ssoins-1234',
            },
        )

        self.account_and_role_prompt = patch.object(
            ConfigureSSOCommand,
            '_prompt_for_sso_account_and_role',
            return_value=('123456789012', 'role'),
        )

        self.region_prompt = patch.object(
            ConfigureSSOCommand,
            '_prompt_for_cli_default_region',
            return_value='us-west-2',
        )

        self.output_prompt = patch.object(
            ConfigureSSOCommand,
            '_prompt_for_cli_output_format',
            return_value='json',
        )

        self.profile_prompt = patch.object(
            ConfigureSSOCommand,
            '_prompt_for_profile',
            return_value='my-profile',
        )

        self.registration_args_prompt.start()
        self.account_and_role_prompt.start()
        self.region_prompt.start()
        self.output_prompt.start()
        self.profile_prompt.start()

    def tearDown(self):
        super().tearDown()
        self.registration_args_prompt.stop()
        self.account_and_role_prompt.stop()
        self.region_prompt.stop()
        self.output_prompt.stop()
        self.profile_prompt.stop()

    def test_configure_sso_implicit_auth(self):
        self.add_oidc_auth_code_responses(
            self.access_token,
            include_register_response=True,
        )
        self.run_cmd('configure sso')

    def test_configure_sso_explicit_device(self):
        self.add_oidc_device_responses(
            self.access_token,
            include_register_response=True,
        )
        self.run_cmd('configure sso --use-device-code')

    def test_configure_sso_implicit_device(self):
        self.add_oidc_device_responses(
            self.access_token,
            include_register_response=True,
        )

        # The lack of a session_name should trigger the device flow
        with patch.object(
            ConfigureSSOCommand,
            '_prompt_for_sso_registration_args',
            return_value={
                'session_name': None,
                'sso_region': 'us-east-1',
                'start_url': 'https://identitycenter.amazonaws.com/ssoins-1234',
            },
        ):
            self.run_cmd('configure sso')

    @patch('botocore.session.Session.create_client')
    def test_configure_sso_custom_ca(self, create_client_patch):
        # Profile and mock responses aren't wired up so this fails with 255,
        # but it does get far enough to verify that the internal SSO-OIDC
        # client is created with the custom CA bundle
        self.run_cmd(
            'configure sso --ca-bundle /path/to/ca/bundle.pem',
            expected_rc=255,
        )

        create_client_patch.assert_called_once_with(
            'sso-oidc',
            config=mock.ANY,
            verify='/path/to/ca/bundle.pem',
        )

    @patch('botocore.session.Session.create_client')
    def test_configure_sso_no_verify_ssl(self, create_client_patch):
        # Profile and mock responses aren't wired up so this fails with 255,
        # but it does get far enough to verify that the internal SSO-OIDC
        # client is created with verify=False
        self.run_cmd(
            'configure sso --no-verify-ssl',
            expected_rc=255,
        )

        create_client_patch.assert_called_once_with(
            'sso-oidc',
            config=mock.ANY,
            verify=False,
        )



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import datetime
import errno
import logging
import os
import shutil
import tempfile

import pytest
from dateutil.tz import tzlocal
from s3transfer.checksums import FullObjectChecksum
from s3transfer.crt import CRTTransferFuture, CRTTransferMeta
from s3transfer.futures import TransferFuture, TransferMeta
from s3transfer.manager import TransferConfig

from awscli.compat import queue
from awscli.customizations.s3 import utils
from awscli.customizations.s3.fileinfo import FileInfo
from awscli.customizations.s3.results import WarningResult
from awscli.customizations.s3.subscribers import (
    CaseConflictCleanupSubscriber,
    CopyPropsSubscriberFactory,
    CreateDirectoryError,
    DeleteCopySourceObjectSubscriber,
    DeleteSourceFileSubscriber,
    DeleteSourceObjectSubscriber,
    DirectoryCreatorSubscriber,
    OnDoneFilteredSubscriber,
    ProvideETagSubscriber,
    ProvideFullObjectChecksumSubscriber,
    ProvideLastModifiedTimeSubscriber,
    ProvideSizeSubscriber,
    ProvideUploadContentTypeSubscriber,
    ReplaceMetadataDirectiveSubscriber,
    ReplaceTaggingDirectiveSubscriber,
    SetMetadataDirectivePropsSubscriber,
    SetTagsSubscriber,
)
from awscli.testutils import FileCreator, mock, unittest
from tests.unit.customizations.s3 import (
    FakeTransferFuture,
    FakeTransferFutureCallArgs,
    FakeTransferFutureMeta,
)


class TestProvideSizeSubscriber(unittest.TestCase):
    def setUp(self):
        self.transfer_future = mock.Mock(spec=TransferFuture)
        self.transfer_meta = TransferMeta()
        self.transfer_future.meta = self.transfer_meta

    def test_size_set(self):
        self.transfer_meta.provide_transfer_size(5)
        subscriber = ProvideSizeSubscriber(10)
        subscriber.on_queued(self.transfer_future)
        self.assertEqual(self.transfer_meta.size, 10)

    def test_does_not_try_to_set_size_on_crt_transfer_future(self):
        crt_transfer_future = mock.Mock(spec=CRTTransferFuture)
        crt_transfer_future.meta = CRTTransferMeta()
        subscriber = ProvideSizeSubscriber(10)
        try:
            subscriber.on_queued(crt_transfer_future)
        except AttributeError:
            self.fail(
                'Subscriber should not have used provide_transfer_size '
                'method because it is not available'
            )


class TestProvideEtagSubscriber:
    def test_etag_set(self):
        transfer_meta = TransferMeta()
        transfer_future = mock.Mock(spec=TransferFuture)
        transfer_future.meta = transfer_meta
        etag = 'myetag'

        transfer_meta.provide_object_etag('oldetag')
        subscriber = ProvideETagSubscriber(etag)
        subscriber.on_queued(transfer_future)
        assert transfer_meta.etag == etag

    def test_does_not_try_to_set_etag_on_crt_transfer_future(self, caplog):
        caplog.set_level(logging.DEBUG)
        crt_transfer_future = mock.Mock(spec=CRTTransferFuture)
        crt_transfer_future.meta = CRTTransferMeta()

        subscriber = ProvideETagSubscriber('myetag')
        subscriber.on_queued(crt_transfer_future)
        assert not hasattr(crt_transfer_future.meta, 'etag')

        assert "Not providing object ETag." in caplog.text


class TestProvideFullObjectChecksumSubscriber:
    def test_checksum_set(self):
        transfer_meta = TransferMeta()
        transfer_future = mock.Mock(spec=TransferFuture)
        transfer_future.meta = transfer_meta
        checksum_info = FullObjectChecksum('crc32', 'abc123==')

        subscriber = ProvideFullObjectChecksumSubscriber(checksum_info)
        subscriber.on_queued(transfer_future)
        assert transfer_meta.full_object_checksum == checksum_info

    def test_does_not_try_to_set_on_crt_transfer_future(self, caplog):
        caplog.set_level(logging.DEBUG)
        crt_transfer_future = mock.Mock(spec=CRTTransferFuture)
        crt_transfer_future.meta = CRTTransferMeta()

        subscriber = ProvideFullObjectChecksumSubscriber(
            FullObjectChecksum('crc32', 'abc123==')
        )
        subscriber.on_queued(crt_transfer_future)
        assert not hasattr(crt_transfer_future.meta, 'full_object_checksum')
        assert "Not providing full object checksum." in caplog.text


class OnDoneFilteredRecordingSubscriber(OnDoneFilteredSubscriber):
    def __init__(self):
        self.on_success_calls = []
        self.on_failure_calls = []

    def _on_success(self, future):
        self.on_success_calls.append(future)

    def _on_failure(self, future, exception):
        self.on_failure_calls.append((future, exception))


class TestOnDoneFilteredSubscriber(unittest.TestCase):
    def test_on_success(self):
        subscriber = OnDoneFilteredRecordingSubscriber()
        future = FakeTransferFuture('return-value')
        subscriber.on_done(future)
        self.assertEqual(subscriber.on_success_calls, [future])
        self.assertEqual(subscriber.on_failure_calls, [])

    def test_on_failure(self):
        subscriber = OnDoneFilteredRecordingSubscriber()
        exception = Exception('my exception')
        future = FakeTransferFuture(exception=exception)
        subscriber.on_done(future)
        self.assertEqual(subscriber.on_failure_calls, [(future, exception)])
        self.assertEqual(subscriber.on_success_calls, [])


class TestProvideUploadContentTypeSubscriber(unittest.TestCase):
    def setUp(self):
        self.filename = 'myfile.txt'
        self.extra_args = {}
        self.future = self.set_future()
        self.subscriber = ProvideUploadContentTypeSubscriber()

    def set_future(self):
        call_args = FakeTransferFutureCallArgs(
            fileobj=self.filename, extra_args=self.extra_args
        )
        meta = FakeTransferFutureMeta(call_args=call_args)
        return FakeTransferFuture(meta=meta)

    def test_on_queued_provides_content_type(self):
        self.subscriber.on_queued(self.future)
        self.assertEqual(self.extra_args, {'ContentType': 'text/plain'})

    def test_on_queued_does_not_provide_content_type_when_unknown(self):
        self.filename = 'file-with-no-extension'
        self.future = self.set_future()
        self.subscriber.on_queued(self.future)
        self.assertEqual(self.extra_args, {})


class BaseTestWithFileCreator(unittest.TestCase):
    def setUp(self):
        self.file_creator = FileCreator()

    def tearDown(self):
        self.file_creator.remove_all()


class TestProvideLastModifiedTimeSubscriber(BaseTestWithFileCreator):
    def setUp(self):
        super(TestProvideLastModifiedTimeSubscriber, self).setUp()
        self.filename = self.file_creator.create_file('myfile', 'my contents')
        self.desired_utime = datetime.datetime(
            2016, 1, 18, 7, 0, 0, tzinfo=tzlocal()
        )
        self.result_queue = queue.Queue()
        self.subscriber = ProvideLastModifiedTimeSubscriber(
            self.desired_utime, self.result_queue
        )

        call_args = FakeTransferFutureCallArgs(fileobj=self.filename)
        meta = FakeTransferFutureMeta(call_args=call_args)
        self.future = FakeTransferFuture(meta=meta)

    def test_on_success_modifies_utime(self):
        self.subscriber.on_done(self.future)
        _, utime = utils.get_file_stat(self.filename)
        self.assertEqual(utime, self.desired_utime)

    def test_on_success_failure_in_utime_mod_raises_warning(self):
        self.subscriber = ProvideLastModifiedTimeSubscriber(
            None, self.result_queue
        )
        self.subscriber.on_done(self.future)
        # Because the time to provide was None it will throw an exception
        # which results in the a warning about the utime not being able
        # to be set being placed in the result queue.
        result = self.result_queue.get()
        self.assertIsInstance(result, WarningResult)
        self.assertIn(
            'unable to update the last modified time', result.message
        )


class TestDirectoryCreatorSubscriber(BaseTestWithFileCreator):
    def setUp(self):
        super(TestDirectoryCreatorSubscriber, self).setUp()
        self.directory_to_create = os.path.join(
            self.file_creator.rootdir, 'new-directory'
        )
        self.filename = os.path.join(self.directory_to_create, 'myfile')

        call_args = FakeTransferFutureCallArgs(fileobj=self.filename)
        meta = FakeTransferFutureMeta(call_args=call_args)
        self.future = FakeTransferFuture(meta=meta)

        self.subscriber = DirectoryCreatorSubscriber()

    def test_on_queued_creates_directories_if_do_not_exist(self):
        self.subscriber.on_queued(self.future)
        self.assertTrue(os.path.exists(self.directory_to_create))

    def test_on_queued_does_not_create_directories_if_exist(self):
        os.makedirs(self.directory_to_create)
        # This should not cause any issues if the directory already exists
        self.subscriber.on_queued(self.future)
        # The directory should still exist
        self.assertTrue(os.path.exists(self.directory_to_create))

    def test_on_queued_failure_propogates_create_directory_error(self):
        # If makedirs() raises an OSError of exception, we should
        # propogate the exception with a better worded CreateDirectoryError.
        with mock.patch('os.makedirs') as makedirs_patch:
            makedirs_patch.side_effect = OSError()
            with self.assertRaises(CreateDirectoryError):
                self.subscriber.on_queued(self.future)
        self.assertFalse(os.path.exists(self.directory_to_create))

    def test_on_queued_failure_propogates_clear_error_message(self):
        # If makedirs() raises an OSError of exception, we should
        # propogate the exception.
        with mock.patch('os.makedirs') as makedirs_patch:
            os_error = OSError()
            os_error.errno = errno.EEXIST
            makedirs_patch.side_effect = os_error
            # The on_queued should not raise an error if the directory
            # already exists
            try:
                self.subscriber.on_queued(self.future)
            except Exception as e:
                self.fail(
                    'on_queued should not have raised an exception related '
                    'to directory creation especially if one already existed '
                    'but got %s' % e
                )


class TestDeleteSourceObjectSubscriber(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.bucket = 'mybucket'
        self.key = 'mykey'
        call_args = FakeTransferFutureCallArgs(
            bucket=self.bucket, key=self.key, extra_args={}
        )
        meta = FakeTransferFutureMeta(call_args=call_args)
        self.future = mock.Mock()
        self.future.meta = meta

    def test_deletes_object(self):
        DeleteSourceObjectSubscriber(self.client).on_done(self.future)
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key
        )
        self.future.set_exception.assert_not_called()

    def test_sets_exception_on_error(self):
        exception = ValueError()
        self.client.delete_object.side_effect = exception
        DeleteSourceObjectSubscriber(self.client).on_done(self.future)
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key
        )
        self.future.set_exception.assert_called_once_with(exception)

    def test_with_request_payer(self):
        self.future.meta.call_args.extra_args = {'RequestPayer': 'requester'}
        DeleteSourceObjectSubscriber(self.client).on_done(self.future)
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key, RequestPayer='requester'
        )


class TestDeleteCopySourceObjectSubscriber(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.bucket = 'mybucket'
        self.key = 'mykey'
        copy_source = {'Bucket': self.bucket, 'Key': self.key}
        call_args = FakeTransferFutureCallArgs(
            copy_source=copy_source, extra_args={}
        )
        meta = FakeTransferFutureMeta(call_args=call_args)
        self.future = mock.Mock()
        self.future.meta = meta

    def test_deletes_object(self):
        DeleteCopySourceObjectSubscriber(self.client).on_done(self.future)
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key
        )
        self.future.set_exception.assert_not_called()

    def test_sets_exception_on_error(self):
        exception = ValueError()
        self.client.delete_object.side_effect = exception
        DeleteCopySourceObjectSubscriber(self.client).on_done(self.future)
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key
        )
        self.future.set_exception.assert_called_once_with(exception)

    def test_with_request_payer(self):
        self.future.meta.call_args.extra_args = {'RequestPayer': 'requester'}
        DeleteCopySourceObjectSubscriber(self.client).on_done(self.future)
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key, RequestPayer='requester'
        )


class TestDeleteSourceFileSubscriber(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.filename = os.path.join(self.tempdir, 'myfile')
        call_args = FakeTransferFutureCallArgs(fileobj=self.filename)
        meta = FakeTransferFutureMeta(call_args=call_args)
        self.future = mock.Mock()
        self.future.meta = meta

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_deletes_file(self):
        with open(self.filename, 'w') as f:
            f.write('data')
        DeleteSourceFileSubscriber().on_done(self.future)
        self.assertFalse(os.path.exists(self.filename))
        self.future.set_exception.assert_not_called()

    def test_sets_exception_on_error(self):
        DeleteSourceFileSubscriber().on_done(self.future)
        self.assertFalse(os.path.exists(self.filename))
        call_args = self.future.set_exception.call_args[0]
        self.assertIsInstance(call_args[0], EnvironmentError)


class BaseCopyPropsSubscriberTest(unittest.TestCase):
    def setUp(self):
        self.client = mock.Mock()
        self.transfer_config = TransferConfig()
        self.cli_params = {}
        self.source_bucket = 'source-bucket'
        self.source_key = 'source-key'
        self.bucket = 'bucket'
        self.key = 'key'
        self.future = self.get_transfer_future()

    def get_transfer_future(self, size=0):
        return FakeTransferFuture(
            meta=FakeTransferFutureMeta(
                call_args=FakeTransferFutureCallArgs(
                    copy_source={
                        'Bucket': self.source_bucket,
                        'Key': self.source_key,
                    },
                    bucket=self.bucket,
                    key=self.key,
                    extra_args={},
                ),
                size=size,
                user_context={},
            )
        )

    def assert_extra_args(self, future, expected_extra_args):
        self.assertEqual(expected_extra_args, future.meta.call_args.extra_args)

    def set_size_for_mp_copy(self, future):
        future.meta.size = 10 * (1024**2)

    def set_cli_params_to_recursive_copy(self):
        self.cli_params['dir_op'] = True


class TestCopyPropsSubscriberFactory(BaseCopyPropsSubscriberTest):
    def setUp(self):
        super(TestCopyPropsSubscriberFactory, self).setUp()
        self.set_cli_params_to_recursive_copy()
        self.factory = CopyPropsSubscriberFactory(
            self.client, self.transfer_config, self.cli_params
        )
        self.fileinfo = self.get_fileinfo()

    def set_copy_props(self, value):
        self.cli_params['copy_props'] = value

    def get_fileinfo(self, **override_kwargs):
        fileinfo_kwargs = {'src': 'src'}
        fileinfo_kwargs.update(override_kwargs)
        return FileInfo(**fileinfo_kwargs)

    def assert_subscriber_classes(self, actual_subscribers, expected_classes):
        self.assertEqual(len(actual_subscribers), len(expected_classes))
        for i, subscriber in enumerate(actual_subscribers):
            self.assertIsInstance(subscriber, expected_classes[i])

    def test_get_subscribers_for_copy_props_none(self):
        self.set_copy_props('none')
        subscribers = self.factory.get_subscribers(self.fileinfo)
        self.assert_subscriber_classes(
            subscribers,
            [
                ReplaceMetadataDirectiveSubscriber,
                ReplaceTaggingDirectiveSubscriber,
            ],
        )

    def test_get_subscribers_for_copy_props_metadata_directive(self):
        self.set_copy_props('metadata-directive')
        subscribers = self.factory.get_subscribers(self.fileinfo)
        self.assert_subscriber_classes(
            subscribers,
            [
                SetMetadataDirectivePropsSubscriber,
                ReplaceTaggingDirectiveSubscriber,
            ],
        )

    def test_get_subscribers_for_copy_props_default(self):
        self.set_copy_props('default')
        subscribers = self.factory.get_subscribers(self.fileinfo)
        self.assert_subscriber_classes(
            subscribers,
            [
                SetMetadataDirectivePropsSubscriber,
                SetTagsSubscriber,
            ],
        )

    def test_get_subscribers_injects_cached_head_object_response(self):
        self.set_copy_props('default')
        self.cli_params['dir_op'] = False
        self.fileinfo.associated_response_data = {'ContentType': 'from-cache'}
        metadata_subscriber = self.factory.get_subscribers(self.fileinfo)[0]
        self.set_size_for_mp_copy(self.future)
        metadata_subscriber.on_queued(self.future)
        self.assertFalse(self.client.head_object.called)
        self.assert_extra_args(self.future, {'ContentType': 'from-cache'})


class TestReplaceMetadataDirectiveSubscriber(BaseCopyPropsSubscriberTest):
    def test_on_queued(self):
        ReplaceMetadataDirectiveSubscriber().on_queued(self.future)
        self.assert_extra_args(self.future, {'MetadataDirective': 'REPLACE'})


class TestReplaceTaggingDirectiveSubscriber(BaseCopyPropsSubscriberTest):
    def test_on_queued(self):
        ReplaceTaggingDirectiveSubscriber().on_queued(self.future)
        self.assert_extra_args(self.future, {'TaggingDirective': 'REPLACE'})


class TestSetMetadataDirectivePropsSubscriber(BaseCopyPropsSubscriberTest):
    def setUp(self):
        super(TestSetMetadataDirectivePropsSubscriber, self).setUp()
        self.head_object_response = {}
        self.subscriber = SetMetadataDirectivePropsSubscriber(
            client=self.client,
            transfer_config=self.transfer_config,
            cli_params=self.cli_params,
            head_object_response=self.head_object_response,
        )
        self.all_props = {
            'CacheControl': 'cache-control',
            'ContentDisposition': 'content-disposition',
            'ContentEncoding': 'content-encoding',
            'ContentLanguage': 'content-language',
            'ContentType': 'content-type',
            'Expires': 'Tue, 07 Jan 2020 20:40:03 GMT',
            'Metadata': {'key': 'value'},
        }

    def set_head_object_response_props(self, **props):
        self.head_object_response.update(props)

    def set_extra_args(self, future, **extra_args):
        future.meta.call_args.extra_args.update(**extra_args)

    def test_does_not_set_props_for_copy_object(self):
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(self.future, {})

    def test_sets_props_for_mp_copy(self):
        self.set_head_object_response_props(**self.all_props)
        self.set_size_for_mp_copy(self.future)
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(self.future, self.all_props)

    def test_filters_out_props_not_in_metadata_directive(self):
        self.set_head_object_response_props(DoNotInclude='do-not-include')
        self.set_size_for_mp_copy(self.future)
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(self.future, {})

    def test_sets_props_for_cache_control_override(self):
        self.set_extra_args(self.future, CacheControl='override')
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'CacheControl': 'override'},
        )

    def test_sets_props_for_content_disposition_override(self):
        self.set_extra_args(self.future, ContentDisposition='override')
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'ContentDisposition': 'override'},
        )

    def test_sets_props_for_content_encoding_override(self):
        self.set_extra_args(self.future, ContentEncoding='override')
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'ContentEncoding': 'override'},
        )

    def test_sets_props_for_content_language_override(self):
        self.set_extra_args(self.future, ContentLanguage='override')
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'ContentLanguage': 'override'},
        )

    def test_sets_props_for_content_type_override(self):
        self.set_extra_args(self.future, ContentType='override')
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'ContentType': 'override'},
        )

    def test_sets_props_for_expires_override(self):
        self.set_extra_args(self.future, Expires='override')
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'Expires': 'override'},
        )

    def test_sets_props_for_metadata_override(self):
        self.set_extra_args(self.future, Metadata={'key': 'override'})
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(
            self.future,
            {'MetadataDirective': 'REPLACE', 'Metadata': {'key': 'override'}},
        )

    def test_overrides_merges_with_head_object_data(self):
        self.set_head_object_response_props(**self.all_props)
        self.set_extra_args(
            self.future, ContentType='override', Metadata={'key': 'override'}
        )
        self.subscriber.on_queued(self.future)
        expected_extra_args = {
            'CacheControl': 'cache-control',
            'ContentDisposition': 'content-disposition',
            'ContentEncoding': 'content-encoding',
            'ContentLanguage': 'content-language',
            'ContentType': 'override',
            'Expires': 'Tue, 07 Jan 2020 20:40:03 GMT',
            'Metadata': {'key': 'override'},
            'MetadataDirective': 'REPLACE',
        }
        self.assert_extra_args(self.future, expected_extra_args)

    def test_can_override_all_metadata_props(self):
        self.set_head_object_response_props(**self.all_props)
        all_override_args = {
            'CacheControl': 'override',
            'ContentDisposition': 'override',
            'ContentEncoding': 'override',
            'ContentLanguage': 'override',
            'ContentType': 'override',
            'Expires': 'override',
            'Metadata': {'key': 'override'},
        }
        self.set_extra_args(self.future, **all_override_args)
        self.subscriber.on_queued(self.future)
        expected_args = all_override_args.copy()
        expected_args['MetadataDirective'] = 'REPLACE'
        self.assert_extra_args(self.future, expected_args)

    def test_makes_head_object_call_if_not_cached(self):
        subscriber = SetMetadataDirectivePropsSubscriber(
            client=self.client,
            transfer_config=self.transfer_config,
            cli_params=self.cli_params,
            head_object_response=None,
        )
        self.client.head_object.return_value = {}
        self.set_size_for_mp_copy(self.future)
        subscriber.on_queued(self.future)
        self.client.head_object.assert_called_with(
            Bucket=self.source_bucket, Key=self.source_key
        )

    def test_add_extra_params_to_head_object_call(self):
        subscriber = SetMetadataDirectivePropsSubscriber(
            client=self.client,
            transfer_config=self.transfer_config,
            cli_params=self.cli_params,
            head_object_response=None,
        )
        self.client.head_object.return_value = {}
        self.cli_params['request_payer'] = 'requester'
        self.set_size_for_mp_copy(self.future)
        subscriber.on_queued(self.future)
        self.client.head_object.assert_called_with(
            Bucket=self.source_bucket,
            Key=self.source_key,
            RequestPayer='requester',
        )

    def test_head_object_doesnt_use_sse_c_destination_key(self):
        subscriber = SetMetadataDirectivePropsSubscriber(
            client=self.client,
            transfer_config=self.transfer_config,
            cli_params={
                'sse_c': 'AES256',
                'sse_c_key': 'destination-key',
            },
            head_object_response=None,
        )
        self.client.head_object.return_value = {}
        self.set_size_for_mp_copy(self.future)
        subscriber.on_queued(self.future)
        self.client.head_object.assert_called_with(
            Bucket=self.source_bucket,
            Key=self.source_key,
        )

    def test_head_object_uses_sse_c_copy_source_key(self):
        subscriber = SetMetadataDirectivePropsSubscriber(
            client=self.client,
            transfer_config=self.transfer_config,
            cli_params={
                'sse_c_copy_source': 'AES256',
                'sse_c_copy_source_key': 'source-key',
                'sse_c': 'AES256',
                'sse_c_key': 'destination-key',
            },
            head_object_response=None,
        )
        self.client.head_object.return_value = {}
        self.set_size_for_mp_copy(self.future)
        subscriber.on_queued(self.future)
        self.client.head_object.assert_called_with(
            Bucket=self.source_bucket,
            Key=self.source_key,
            SSECustomerAlgorithm='AES256',
            SSECustomerKey='source-key',
        )


class PutObjectTaggingException(Exception):
    pass


class TestSetTagsSubscriber(BaseCopyPropsSubscriberTest):
    def setUp(self):
        super(TestSetTagsSubscriber, self).setUp()
        self.source_client = mock.Mock()
        self.subscriber = SetTagsSubscriber(
            client=self.client,
            transfer_config=self.transfer_config,
            cli_params=self.cli_params,
            source_client=self.source_client,
        )
        self.tagging_response = {
            'TagSet': [
                {'Key': 'tag1', 'Value': 'val1'},
                {'Key': 'tag2', 'Value': 'val2'},
            ]
        }
        self.url_encoded_tags = 'tag1=val1&tag2=val2'
        self.tagging_response_over_limit = {
            'TagSet': [
                {'Key': 'tag', 'Value': 'val1' * (2 * 1024)},
            ]
        }

    def test_does_not_set_tags_for_copy_object(self):
        self.subscriber.on_queued(self.future)
        self.assert_extra_args(self.future, {})

    def test_sets_tags_for_mp_copy(self):
        self.set_size_for_mp_copy(self.future)
        self.source_client.get_object_tagging.return_value = (
            self.tagging_response
        )
        self.subscriber.on_queued(self.future)
        self.source_client.get_object_tagging.assert_called_with(
            Bucket=self.source_bucket, Key=self.source_key
        )
        self.assert_extra_args(self.future, {'Tagging': self.url_encoded_tags})

    def test_does_not_set_tags_if_not_present(self):
        self.set_size_for_mp_copy(self.future)
        self.source_client.get_object_tagging.return_value = {'TagSet': []}
        self.subscriber.on_queued(self.future)
        self.source_client.get_object_tagging.assert_called_with(
            Bucket=self.source_bucket, Key=self.source_key
        )
        self.assert_extra_args(self.future, {})

    def test_sets_tags_using_put_object_tagging_if_over_size_limit(self):
        self.set_size_for_mp_copy(self.future)
        self.source_client.get_object_tagging.return_value = (
            self.tagging_response_over_limit
        )
        self.subscriber.on_queued(self.future)
        self.source_client.get_object_tagging.assert_called_with(
            Bucket=self.source_bucket, Key=self.source_key
        )
        self.assert_extra_args(self.future, {})
        self.subscriber.on_done(self.future)
        self.client.put_object_tagging.assert_called_with(
            Bucket=self.bucket,
            Key=self.key,
            Tagging=self.tagging_response_over_limit,
        )

    def test_does_not_call_put_object_tagging_if_transfer_fails(self):
        self.set_size_for_mp_copy(self.future)
        self.source_client.get_object_tagging.return_value = (
            self.tagging_response_over_limit
        )
        self.subscriber.on_queued(self.future)
        self.future.set_exception(Exception())
        self.subscriber.on_done(self.future)
        self.assertFalse(self.client.put_object_tagging.called)

    def test_put_object_tagging_propagates_error_and_cleans_up_if_fails(self):
        self.set_size_for_mp_copy(self.future)
        self.source_client.get_object_tagging.return_value = (
            self.tagging_response_over_limit
        )
        self.subscriber.on_queued(self.future)

        self.client.put_object_tagging.side_effect = (
            PutObjectTaggingException()
        )
        self.subscriber.on_done(self.future)

        with self.assertRaises(PutObjectTaggingException):
            self.future.result()

        self.client.put_object_tagging.assert_called_with(
            Bucket=self.bucket,
            Key=self.key,
            Tagging=self.tagging_response_over_limit,
        )
        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket,
            Key=self.key,
        )

    def test_add_extra_params_to_delete_object_call(self):
        self.cli_params['request_payer'] = 'requester'
        self.set_size_for_mp_copy(self.future)
        self.source_client.get_object_tagging.return_value = (
            self.tagging_response_over_limit
        )
        self.subscriber.on_queued(self.future)

        self.client.put_object_tagging.side_effect = (
            PutObjectTaggingException()
        )
        self.subscriber.on_done(self.future)

        self.client.delete_object.assert_called_once_with(
            Bucket=self.bucket, Key=self.key, RequestPayer='requester'
        )


class TestCaseConflictCleanupSubscriber:
    def test_on_done_removes_key_from_set(self):
        submitted = {123, 456}
        subscriber = CaseConflictCleanupSubscriber(submitted, 123)
        subscriber.on_done(future=None)
        assert submitted == {456}

    def test_on_done_handles_missing_key(self):
        submitted = {456}
        subscriber = CaseConflictCleanupSubscriber(submitted, 123)
        subscriber.on_done(future=None)
        assert submitted == {456}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestDDB(BaseAWSCommandParamsTest):
    def test_no_command_specified(self):
        _, stderr, _ = self.run_cmd('ddb', expected_rc=252)
        expected = (
            "usage: aws [options] ddb <subcommand> "
            "[parameters]\naws: [ERROR]: too few arguments"
        )
        self.assertIn(expected, stderr)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from datetime import datetime

from dateutil import tz

from awscli.testutils import BaseAWSCommandParamsTest, mock


class TestTailCommand(BaseAWSCommandParamsTest):
    def setUp(self):
        super(TestTailCommand, self).setUp()
        self.group_name = 'mygroup'
        self.stream_name = 'mystream'
        self.response_log_timestamp = 1100  # 1.1 seconds
        self.formatted_log_timestamp = '1970-01-01T00:00:01.100000+00:00'
        self.formatted_short_log_timestamp = '1970-01-01T00:00:01'
        self.message = 'my message\n'
        self.parsed_responses = [
            {
                "events": [
                    {
                        'eventId': 'event1',
                        'logStreamName': self.stream_name,
                        'message': self.message,
                        'timestamp': self.response_log_timestamp,
                        'ingestionTime': self.response_log_timestamp,
                    }
                ]
            }
        ]

    def tearDown(self):
        super(TestTailCommand, self).tearDown()

    def test_tail(self):
        stdout, _, _ = self.assert_params_for_cmd(
            'logs tail %s' % self.group_name,
            params={
                'logGroupName': self.group_name,
                'interleaved': True,
                'startTime': mock.ANY,
            },
        )
        self.assertEqual(
            stdout,
            '%s %s %s'
            % (self.formatted_log_timestamp, self.stream_name, self.message),
        )

    def test_tail_paginated(self):
        self.parsed_responses = [
            {
                "events": [
                    {
                        'eventId': 'event1',
                        'logStreamName': self.stream_name,
                        'message': self.message,
                        'timestamp': self.response_log_timestamp,
                        'ingestionTime': self.response_log_timestamp,
                    }
                ],
                'nextToken': 'token',
            },
            {
                "events": [
                    {
                        'eventId': 'event2',
                        'logStreamName': self.stream_name,
                        'message': self.message,
                        'timestamp': self.response_log_timestamp,
                        'ingestionTime': self.response_log_timestamp,
                    }
                ],
            },
        ]
        stdout, _, _ = self.run_cmd('logs tail %s' % self.group_name)
        self.assertEqual(len(self.operations_called), 2)
        self.assertEqual(self.operations_called[0][0].name, 'FilterLogEvents')
        self.assertEqual(
            self.operations_called[0][1],
            {
                'logGroupName': 'mygroup',
                'interleaved': True,
                'startTime': mock.ANY,
            },
        )
        self.assertEqual(self.operations_called[1][0].name, 'FilterLogEvents')
        self.assertEqual(
            self.operations_called[1][1],
            {
                'logGroupName': 'mygroup',
                'interleaved': True,
                'startTime': mock.ANY,
                'nextToken': 'token',
            },
        )
        self.assertEqual(
            stdout,
            '%s %s %s'
            % (self.formatted_log_timestamp, self.stream_name, self.message)
            * 2,
        )

    def test_tail_with_follow(self):
        self.parsed_responses = [
            {
                "events": [
                    {
                        'eventId': 'event1',
                        'logStreamName': self.stream_name,
                        'message': self.message,
                        'timestamp': self.response_log_timestamp,
                        'ingestionTime': self.response_log_timestamp,
                    }
                ],
            },
            {
                "events": [
                    {
                        'eventId': 'event2',
                        'logStreamName': self.stream_name,
                        'message': self.message,
                        'timestamp': self.response_log_timestamp,
                        'ingestionTime': self.response_log_timestamp,
                    }
                ],
            },
        ]
        with mock.patch('time.sleep') as mock_sleep:
            mock_sleep.side_effect = [None, KeyboardInterrupt]
            stdout, _, _ = self.run_cmd(
                'logs tail %s --follow' % self.group_name
            )
            self.assertEqual(
                stdout,
                '%s %s %s'
                % (
                    self.formatted_log_timestamp,
                    self.stream_name,
                    self.message,
                )
                * 2,
            )

    def test_tail_defaults_to_10m(self):
        datetime_mock = mock.Mock(wraps=datetime)
        datetime_mock.utcnow = mock.Mock(
            return_value=datetime(1970, 1, 1, 0, 10, 1, tzinfo=tz.tzutc())
        )
        with mock.patch(
            'awscli.customizations.logs.tail.datetime', new=datetime_mock
        ):
            self.assert_params_for_cmd(
                'logs tail %s' % self.group_name,
                params={
                    'logGroupName': self.group_name,
                    'interleaved': True,
                    'startTime': 1000,
                },
            )

    def test_tail_with_since(self):
        self.assert_params_for_cmd(
            'logs tail %s --since 1970-01-01T00:00:01' % self.group_name,
            params={
                'logGroupName': self.group_name,
                'interleaved': True,
                'startTime': 1000,
            },
        )

    def test_tail_with_relative_since(self):
        datetime_mock = mock.Mock(wraps=datetime)
        datetime_mock.utcnow = mock.Mock(
            return_value=datetime(1970, 1, 1, 0, 0, 2, tzinfo=tz.tzutc())
        )
        with mock.patch(
            'awscli.customizations.logs.tail.datetime', new=datetime_mock
        ):
            self.assert_params_for_cmd(
                'logs tail %s --since 1s' % self.group_name,
                params={
                    'logGroupName': self.group_name,
                    'interleaved': True,
                    'startTime': 1000,
                },
            )

    def test_tail_with_filter_pattern(self):
        self.assert_params_for_cmd(
            'logs tail %s --filter-pattern Pattern' % self.group_name,
            params={
                'logGroupName': self.group_name,
                'interleaved': True,
                'startTime': mock.ANY,
                'filterPattern': 'Pattern',
            },
        )

    def test_tail_with_short_format(self):
        stdout, _, _ = self.run_cmd(
            'logs tail %s --format short' % self.group_name
        )
        self.assertEqual(
            stdout,
            '%s %s' % (self.formatted_short_log_timestamp, self.message),
        )

    def test_tail_with_color_on(self):
        stdout, _, _ = self.run_cmd(
            'logs tail %s --color on' % self.group_name
        )
        self.assertEqual(
            stdout,
            "\x1b[32m%s\x1b[0m \x1b[36m%s\x1b[0m %s"
            % (self.formatted_log_timestamp, self.stream_name, self.message),
        )

    def test_tail_with_color_off(self):
        stdout, _, _ = self.run_cmd(
            'logs tail %s --color off' % self.group_name
        )
        self.assertEqual(
            stdout,
            "%s %s %s"
            % (self.formatted_log_timestamp, self.stream_name, self.message),
        )

    def test_tail_no_color_when_tty(self):
        with mock.patch(
            'awscli.customizations.logs.tail.is_a_tty'
        ) as mock_is_a_tty:
            mock_is_a_tty.return_value = True
            stdout, _, _ = self.run_cmd('logs tail %s' % self.group_name)
        self.assertEqual(
            stdout,
            "\x1b[32m%s\x1b[0m \x1b[36m%s\x1b[0m %s"
            % (self.formatted_log_timestamp, self.stream_name, self.message),
        )

    def test_tail_with_log_stream_names(self):
        self.assert_params_for_cmd(
            'logs tail %s --log-stream-names foo-stream bar-stream'
            % self.group_name,
            params={
                'logGroupName': self.group_name,
                'interleaved': True,
                'startTime': mock.ANY,
                'logStreamNames': ['foo-stream', 'bar-stream'],
            },
        )

    def test_tail_with_log_stream_name_prefix(self):
        self.assert_params_for_cmd(
            'logs tail %s --log-stream-name-prefix foo' % self.group_name,
            params={
                'logGroupName': self.group_name,
                'interleaved': True,
                'startTime': mock.ANY,
                'logStreamNamePrefix': 'foo',
            },
        )
